module TiContextReduction(entails,contextReduction,contextReduction'') where
import TiInstanceDB
import TiClasses
import TiMonad
import TiTypes
import TiBySuper
import TiFresh
import TiNames(TypeId,ValueId,dictName)
import TiSCC(sccD')
import TiFreeNames(freeVars)
import TiSolve(expandSynonyms)
import SrcLoc
import PrettyPrint
import List(partition)
import MUtils
import IOExts(trace) -- debug
import TiPretty() -- debug

--type Ctx = [Typing QId Pred]

--contextReduction, rmByInst :: HasLocalDef e => IDB -> Ctx -> IM i (Ctx,e->e)


type Dict v = Typing v (Type v)
type Ctx v = [Dict v]

type DDefs v = [(Dict v,DExp v)] -- dictionary definitions
type DictsF v = DDefs v->DDefs v

contextReduction ds = apSnd returnLets # contextReduction1 ds

contextReduction1 ds =
 do idb <- getIEnv
    tenv <- getKEnv
    contextReduction' idb tenv ds

contextReduction'
  :: (TypeId v,ValueId v,HasSrcLoc v)
  => IDB v -> KEnv v -> Ctx v -> IM i c (Ctx v, DictsF v)
contextReduction' idb env =
    rmDupCtx & rmBySuper & rmDupCtx & rmByInst env idb & rmDupCtx
  where
    (&) red1 red2 ds =
      do (ds1,r1) <- red1 ds
	 (ds2,r2) <- red2 ds1
	 return (ds2,r2 . r1)

    rmDupCtx = return . rmDupCtx'
    rmDupCtx' ds =
      case ds of
	[] -> ([],id)
	(d@(dn:>:dt):ds) ->
	   case partition (sameInstance d) ds of
	     (same,other) -> (d:ds',r')
	       where
		 r' = foldr l r same
		 l d' = (letvar' d' dne .)
		 dne = var dn
	    	 (ds',r) = rmDupCtx' other

    rmBySuper ds =
      do let addSuper (dn:>:dt) = tail # bySuper env (var dn:>:dt)
         supds <- concatMapM addSuper ds
	 let rm ds1 r [] = (reverse ds1,r)
	     rm ds1 r (d:ds2) =
		 case filter (sameInstance d) supds of
		   [] -> rm (d:ds1) r ds2
		   (supde:>:_):_ -> rm ds1 r' ds2
		     where r' = letvar' d supde . r
	 return (rm [] id ds)

entails eds ids env =
    do availds <- concatMapM (bySuper env) (map (emap var) eds)
       returnLets . foldr (.) id # mapM (entails1 availds) ids
  where
    entails1 availds d@(_:>:dt) =
       case filter (sameInstance d) availds of
         [] -> fail (pp $ fsep[ppi "contexts don't match:",
			       ppiFTuple ectx,ppi "vs",ppi dt])
	 (supde:>:_):_ -> return (letvar' d supde)

    _:>:ectx = unzipTyped eds   

sameInstance (_:>:t1) (_:>:t2) = t1==t2

rmByInst env idb ds =
  --stpp ds $
  case ds of
    [] -> return ([],id)
    (d@(dn:>:inst):ds) ->
      case findInst idb (expandSynonyms env inst) of
	[] -> {-(if null (tv inst)
	       then trace (pp $ fsep [ppi (srcLoc dn)<>":",
				      ppi "No instance: ",
				      ppi inst{-,
				      ppi "in",
				      ppi idb-}])
	       else id) $-}
	      preserve
	[i] -> if True -- null (tv inst) -- delay reduction, ...
               then useInst i
	       else trace (pp $ "Delaying"<+>d<+>srcLoc dn) $
                    preserve -- ... experiment for overlapping instances
	is -> trace msg preserve -- experiment
	      --fail msg -- normal
          where msg = pp $ sep [ppi "Ambiguous instances", "("<>map fst is<>")",
		                "for"<+>inst]
      where
        preserve = apFst (d:) # rmByInst env idb ds

	useInst ((idn,ips),((gs,dt),s)) =
	    do idns <- map dictName # freshlist (length ips)
	       let ids = zipWith (:>:) idns ips
		   --Forall gs _ = sc
		   sc = mono dt -- fake, ok since it isn't used
		   ts = apply s (map tyvar gs)
		   --idne = var idn -- without type annotations
		   idne = DSpec (HsVar idn) sc ts -- with type annotations
		   d' = foldl1 DApp (idne:map var idns)
		   r1 = letvar' d d'
	       (ds2,r2) <- rmByInst env idb (ids++ds)
	       return (ds2,r2 . r1)

letvar' v e = ((v,e):)

-- Sort dictionary definitions in dependency order and return a nested let.
returnLets r = \s -> foldr (uncurry letvar . apSnd dconv) s (order (r []))
  where
    order = concat . sccD' names
    names (d:>:_,e) = ([HsVar d],dfree e)

--------------------------------------------------------------------------------

-- To avoid an ambiguity when you only want the resulting predicates and
-- not the dictionary rewriting function:

contextReduction'' ds = fst # contextReduction1 ds

data DExp i
  = DId (HsIdentI i)
  | DApp (DExp i) (DExp i)
  | DSpec (HsIdentI i) (Scheme i) [Type i]

dconv de =
  case de of
    DId i -> ident i
    DApp e1 e2 -> app (dconv e1) (dconv e2)
    DSpec i sc t -> spec i sc t

instance HasId i (DExp i) where ident = DId
instance HasCoreSyntax i (DExp i) where app = DApp -- ; ...
instance HasTypeApp i (DExp i) where spec = DSpec

dfree de =
    case de of
      DId   i     -> [i]
      DSpec i _ _ -> [i]
      DApp  e1 e2 -> dfree e1++dfree e2
