module SimpPatMatch(module SimpPatMatch,HasOrig,HasIdTy,HasSrcLoc,ValueId) where
import Maybe(isJust,fromMaybe,fromJust)
import Substitute
import Recursive
import HasBaseStruct
import BaseSyntax
import DefinedNames(definedVars)
import TiNames(ValueId,localVal')
import TiPrelude(prelError,prelEqual)
import TiClasses(HasDef,noDef,nullDef)
import TypedIds
import UniqueNames(orig,HasOrig)
import FreeNames
--import DefinedNames
import PrettyPrint(pp)

simpAllPatMatch m = mapExp simpPatMatch m

class SimpPatMatch s where
  simpPatMatch :: s -> s

instance SimpPatMatch s => SimpPatMatch [s] where
  simpPatMatch = map simpPatMatch

simpPatMatchE e = simpPatMatchE' rec e

simpPatMatchE' rec e0 =
    case mapEI id simpPatMatch id simpAllPatMatch id id e0 of
      HsCase e alts -> compilePatMatch e alts
      e -> rec e

{-+
Pattern match simplification, as described in section 3.17.3
of the Haskell 98 report.
-}
{-
compilePatMatch ::
   (ValueId i,HasSrcLoc i,HasIdTy n i,HasOrig i,HasOrig n,
    HasBaseStruct e (EI i e p ds t c),GetBaseStruct e (EI i e p ds t c),
    FreeNames i e,DefinedNames i p,
    HasBaseStruct p (PI i p),GetBaseStruct p (PI i p),
    HasBaseStruct d (DI i e p ds t c tp),
    HasDef ds d)
  => e -> [HsAlt e p ds] -> e
--}
-- Rule (a)
compilePatMatch e alts = mustBeVar' "v0" (srcLoc alts) e (patMatch_b alts)

-- Rule (b)
patMatch_b alts v = foldr (patMatch_c v) err alts
  where
    err = nomatch s "No match"
    s = srcLoc alts

nomatch s msg = hsId prelError `hsApp` hsLit s (HsString (pp s++": "++msg))

-- Rule (c)
patMatch_c v (HsAlt s p rhs ds) e' =
    case rhs of
      HsBody e -> patMatch_c' v s p e ds e'
      HsGuard gdrhss -> mustBeVar s e' (rhs.hsEVar)
        where
          rhs ye = patMatch_c' v s p (ifGuards ye gdrhss) ds ye
          ifGuards = foldr ifGuard
          ifGuard (s,g,e) = hsIf g e
  where
    -- Rule (c), after the guards have been handled
    patMatch_c' v s p e ds e' = patMatch_defghij v s p (hsLet' ds e) e'

-- Introduce a variable to avoid code duplication, if necessary:
mustBeVar s = mustBeVar' "y0" s
mustBeVar' n s e cont =
  case isHsIdVar =<< basestruct e of
    Just y -> cont y
    _ -> monoLet y e (cont y)
      where
	y = localVal' n (Just s)

-- Rule (d)-(j)
patMatch_defghij v s p rhs e' =
  case basestruct p of
    Just bp ->
      case bp of
        HsPIrrPat p' -> patMatch_d v s p' rhs -- rule (d)
	HsPAsPat x p' -> patMatch_defghij v s p' (monoLetVar x v rhs) e' -- rule (e)
	HsPWildCard -> rhs -- rule (f)
	HsPApp k ps@(_:_) -> patMatch_g (srcLoc k) v (hsPApp k) ps rhs e' -- rule (g)
	HsPTuple ps -> patMatch_g s v hsPTuple ps rhs e' -- rule (g)
	HsPInfixApp p1 (HsCon k) p2 -> patMatch_g (srcLoc k) v k' [p1,p2] rhs e' -- rule (g)
          where k' [p1,p2] = hsPInfixApp p1 (HsCon k) p2
	HsPLit s lit -> -- rule (h) 
                      hsIf (hsEVar v `eqTest` hsLit s lit) rhs e'
        HsPNeg s lit ->
            hsIf (hsEVar v `eqTest` hsNegApp s (hsLit s lit)) rhs e'
        HsPId (HsVar x) -> monoLetVar x v rhs -- rules (i) and (j)
	HsPParen p' -> patMatch_defghij v s p' rhs e'
        HsPList ps -> patMatch_g s v hsPList ps rhs e' -- rule (g), hmm!!
        HsPRec k lps -> patMatch_mno v k lps rhs e'

	_ -> keep
    _ -> keep
  where
    keep = patMatch_keep v s p rhs e'

    eqTest e1 e2 = hsId prelEqual `hsApp` e1 `hsApp` e2

-- Rule (d)
patMatch_d v s p' rhs =
     monoLets xs (map proj xs) rhs
  where
    xs = map getHSName (definedVars p')
    proj x = patMatch_defghij v s p' (hsEVar x) 
	                             (nomatch s "Irrefutable pattern failed")

-- Rule (g)
patMatch_g s v k ps rhs e' = mustBeVar s e' (patMatch_g' s v k ps rhs.hsEVar)
patMatch_g' s v k ps rhs e' =
    simpleCase v s p' body e'
  where
     p' = k (map hsPVar xs)
     xs = zipWith reuse ps [localVal' ('x':show n) (Just s)|n<-[1..length ps]]
     body = foldr submatch rhs (zip xs ps)
     submatch (x,p) rest = patMatch_defghij x s p rest e'

     reuse p x = fromMaybe x (isPVar p)

-- Rules (m)-(o), labelled fields
patMatch_mno v k lps rhs e' =
  case lps of
    [] -> patMatch_o v k rhs e'
    [HsField f p] -> patMatch_n v k f p rhs e'
    HsField f p:lps' -> -- rule (m)
       mustBeVar (srcLoc k) e' $ \ y ->
       let ey = hsEVar y
       in patMatch_n v k f p (patMatch_mno v k lps' rhs ey) ey

-- Rule (n)
patMatch_n v k f p rhs e' = simpleCase v (srcLoc k) p' rhs e'
  where p' = hsPApp k (map posPat fs)
        posPat f' = if orig f'==orig f then p else hsPWildCard
	fs = confields k

-- Rule (o)
patMatch_o v k rhs e' = simpleCase v (srcLoc k) p' rhs e'
  where p' = hsPApp k (replicate n hsPWildCard)
        n = conarity k

patMatch_keep v = simpleCase v -- remaining cases and unimplemented rules

simpleCase v s p rhs e' =
    hsCase (hsEVar v) alts
  where
    alts = if irrefutable p
	   then [alt1]
	   else [alt1,alt2]
    alt1 = HsAlt s p (HsBody rhs) noDef
    alt2 = default_alt s e'

    default_alt s e' = HsAlt s hsPWildCard (HsBody e') noDef

monoLetVar x y = if x==y then id else monoLet x (hsEVar y)

-- Monomorphic, nonrecursive variants of let x = e in e'
-- (The Haskell report uses two versions)
--monoLet s x e e' = hsCase e [HsAlt s (hsPVar x) (HsBody e') noDef]
monoLet x e = monoLets [x] [e]
monoLets [] [] e' = e'
monoLets xs0 es0 e' = hsParen (hsLambda' [hsPVar x|x<-xs] e') `apps` es
  where
    apps e es = foldl hsApp e es
    -- Omit bindings for variables that aren't used:
    (xs,es) = unzip . filter (keep.fst) $ zip xs0 es0
    keep x = x `elem` fvs
    fvs = [v|(HsVar v,ValueNames)<-freeNames e']

hsLet' ds e = if nullDef ds then e else hsLet ds e
hsLambda' [] e = e
hsLambda' xs e = hsLambda xs e

irrefutable p =
  isJust (isPVar p) || isJust (isIrrPat p) || isWildCardPat p ||
  case isAsPat p of
    Just (_,p) -> irrefutable p
    _ -> False
  ||
  case basestruct p of
    Just (HsPTuple ps) -> all irrefutable ps
    -- generalize to all single-constructor types!
    _ -> False

conarity k= conArity . conInfo $ k
confields k = fromJust . conFields . conInfo $ k

conInfo k =
  case idTy k of
    ConstrOf _ ti -> head [con|con<-constructors ti,orig (conName con)==orig k]
